--
--      Copyright Cypress Semiconductor Corporation, 1995,1996
--	as an unpublished work.
--

--
--	MCOUNTER
--

library ieee;
use ieee.std_logic_1164.all;

package Mcounter_local is
    type bvm		is array (integer range <>,integer range <>) of std_logic;
    type Shift_Type	is (LPM_LOGICAL, LPM_ROTATE, LPM_ARITHMETIC);
    type Repre_Type	is (LPM_SIGNED, LPM_UNSIGNED, LPM_GRAY);
    type Truth_Type	is (LPM_F, LPM_FD, LPM_FR, LPM_FDR);
    type CtDir_Type	is (LPM_NO_DIR, LPM_UP, LPM_DOWN);
    type Arith_Type	is (LPM_NO_TYP, LPM_ADD, LPM_SUB);
    type Fflop_Type	is (LPM_DFF,LPM_TFF);
    type ShDir_Type	is (LPM_LEFT,LPM_RIGHT);
    type goal_type	is (speed,area,combinatorial);
    type port_state_type is (ps_tied_low,ps_tied_high,ps_open,ps_connected);

    signal zero : std_logic := '0' ;
    signal one : std_logic := '1' ;

    function port_state (arg : std_logic) return port_state_type;
    function connects (x : std_logic; i : integer) return integer;
    function max (a,b : natural) return natural;
    function vectorize (a, size : natural) RETURN std_logic_vector;
    function restring (a : string; w : positive) return std_logic_vector;
component dsrff
    port(
	d	: in std_logic;		-- D input to register.
	s	: in std_logic;		-- Set input Async.
	r	: in std_logic;		-- Reset input Async.
	clk	: in std_logic;		-- Positive edge Clock.
	q	: out std_logic		-- Output from register.
    );
end component;
    component cinc
	generic(width	: natural);
	port(	dataA	: in	Bit_Vector;
		sum	: out	Bit_Vector);
    end component;

    component finc
	generic(width	: natural);
	port(	dataA	: in	Bit_Vector;
		cin	: in	Bit;
		sum	: out	Bit_Vector);
    end component;

    component cadd
	generic(width	: natural;
		firstOne: natural;
		lastOne : natural);
	port(	A,K	: in	Bit_Vector;
		sum	: out	Bit_Vector);
    end component;

    component cdec
	generic(width	: natural);
	port(	dataA	: in	Bit_Vector;
		dif	: out	Bit_Vector);
    end component;

    component fdec
	generic(width	: natural);
	port(	dataA	: in	Bit_Vector;
		bin	: in	Bit;
		dif	: out	Bit_Vector);
    end component;

    component csub
	generic(width	: natural;
		firstOne: natural;
		lastOne : natural);
	port(	A,K	: in	Bit_Vector;
		dif	: out	Bit_Vector);
    end component;

    function to_bv (x: std_logic_vector) return Bit_Vector;
    function from_bv (x: Bit_Vector) return std_logic_vector;

end Mcounter_local;

package body Mcounter_local is
    function port_state (arg : std_logic) return port_state_type is
	variable result : port_state_type := ps_connected;
    begin

	result := arg'port_state_att;

	return result;
    end port_state;

    function connects (x : std_logic; i : integer) return integer is
	variable result : integer := i;
    begin
	if (port_state(x) = ps_connected) then
	    result := result - 1;
	end if;
	return (result);
    end connects;

    function max (a,b : natural) return natural is
    begin
	if (a < b) then return b; else return a; end if;
    end max;

    function restring (a : string; w : positive) return std_logic_vector is
	variable result : std_logic_vector (w-1 downto 0) := (others => '1');
	variable indx : natural;
    begin

	if (a'length = 0) then
	    return result;
	elsif (a'length /= w) then
	    Assert (false)
	    Report "RESTRING: Source string length /= target vector width."
	    Severity Failure;
	    return result;
	else
	    indx := 0;
	    for i in a'reverse_range loop
		case a(i) is
		when '1' => result(indx) := '1';
		when '0' => result(indx) := '0';

		when '-' => result(indx) := '-';

		when others => Assert (false)

		    Report "Invalid char. literal (must be '1', '0', or '-').  Defauting to  '0'."

		    Severity Warning;
		    result(indx) := '0';
		end case;
		indx := indx + 1;
	    end loop;
	    return result;
	end if;
    end restring;

    function to_bv (x: std_logic_vector) return Bit_Vector is
	variable result : Bit_Vector(x'range);
    begin
	for i in x'range loop
	    if x(i) = '1' then
		result(i) := '1';
	    else
		result(i) := '0';
	    end if;
	end loop;
	return result;
    end to_bv;

    function from_bv (x: Bit_Vector) return std_logic_vector is
	variable result : std_logic_vector(x'range);
    begin
	for i in x'range loop
	    if x(i) = '1' then
		result(i) := '1';
	    else
		result(i) := '0';
	    end if;
	end loop;
	return result;
    end from_bv;

end Mcounter_local;

library ieee;
use ieee.std_logic_1164.all;

use work.Mcounter_local.all;

entity Mcounter is
    generic(lpm_width	: positive;
	    lpm_direction	: ctdir_type;
	    lpm_avalue	: string;
	    lpm_svalue	: string;
	    lpm_pvalue	: string;
	    lpm_hint	: goal_type);
    port(   data	: in	std_logic_vector((lpm_width-1) downto 0);
	    clock	: in	std_logic;
	    clk_en	: in	std_logic;
	    cnt_en	: in	std_logic;
	    updown	: in	std_logic;
	    q		: out	std_logic_vector((lpm_width-1) downto 0);
	    aset	: in	std_logic;
	    aclr	: in	std_logic;
	    aload	: in	std_logic;
	    sset	: in	std_logic;
	    sclr	: in	std_logic;
	    sload	: in	std_logic;
	    testenab	: in	std_logic;
	    testin	: in	std_logic;
	    testout	: out	std_logic);
end Mcounter;

architecture archlpmstd of Mcounter is

    signal qtmp : std_logic_vector(q'range);
    signal dinput,async,set,clr : std_logic_vector(q'range);
    signal one : Bit := '1';
    signal dataA, sum, qup, qdn : Bit_Vector (q'range);

begin

    async <= restring(lpm_avalue,lpm_width);

    dataA <= TO_BV(qtmp);

    UP: if (((port_state(updown) = ps_tied_high)
	      AND (lpm_direction = lpm_no_dir))
	       OR (lpm_direction = lpm_up)) generate
	SA: if (lpm_hint = area) generate
	    U0: cinc generic map (lpm_width) port map (dataA, sum);
	end generate;
	SS: if (lpm_hint = speed) generate
	    U0: finc generic map (lpm_width) port map (dataA, one, sum);
	end generate;
    end generate;

    DN: if (((port_state(updown) = ps_tied_low)
	      AND (lpm_direction = lpm_no_dir))
	       OR (lpm_direction = lpm_down)) generate
	SA: if (lpm_hint = area) generate
	    U0: cdec generic map (lpm_width) port map (dataA, sum);
	end generate;
	SS: if (lpm_hint = speed) generate
	    U0: fdec generic map (lpm_width) port map (dataA, one, sum);
	end generate;
    end generate;

    UD: if ((port_state(updown) = ps_connected)
		     AND (lpm_direction = lpm_no_dir)) generate
	SA: if (lpm_hint = area) generate
	    U0: cinc generic map (lpm_width) port map (dataA, qup);
	    U1: cdec generic map (lpm_width) port map (dataA, qdn);
	end generate;
	SS: if (lpm_hint = speed) generate
	    U0: finc generic map (lpm_width) port map (dataA, one, qup);
	    U1: fdec generic map (lpm_width) port map (dataA, one, qdn);
	end generate;
	sum <= qup when (updown = '1') else qdn;
    end generate;

    process (sclr,cnt_en,sload,data,sset,testin,testenab,sum,qtmp)
    begin
	if (testenab = '1') then
	    dinput(0) <= testin;
	    if (lpm_width > 1) then
		dinput(q'high downto 1) <= qtmp((q'high - 1) downto 0);
	    end if;
	elsif (sclr = '1' AND clk_en = '1') then
	    dinput <= (others => '0');
	elsif (sset = '1' AND clk_en = '1') then
	    dinput <= restring(lpm_svalue,lpm_width);
	elsif (sload = '1' AND clk_en = '1') then
	    dinput <= data;
	elsif (cnt_en = '1' AND clk_en = '1') then
	    dinput <= FROM_BV(sum);
	else
	    dinput <= qtmp;
	end if;
    end process;

    L0: for i in q'range generate
	set(i) <= (aset AND     async(i)) OR (aload AND     data(i));
	clr(i) <= (aset AND NOT async(i)) OR (aload AND NOT data(i)) OR aclr;

	U1: dsrff port map(dinput(i),set(i),clr(i),clock,qtmp(i));
    end generate;

    q <= qtmp;
    testout <= qtmp(q'high);

end archlpmstd;
