--
--      Copyright Cypress Semiconductor Corporation, 1995,1996
--	as an unpublished work.
--

--
--	MMULT
--

library ieee;
use ieee.std_logic_1164.all;

package Mmult_local is
    type bvm		is array (integer range <>,integer range <>) of std_logic;
    type Shift_Type	is (LPM_LOGICAL, LPM_ROTATE, LPM_ARITHMETIC);
    type Repre_Type	is (LPM_SIGNED, LPM_UNSIGNED, LPM_GRAY);
    type Truth_Type	is (LPM_F, LPM_FD, LPM_FR, LPM_FDR);
    type CtDir_Type	is (LPM_NO_DIR, LPM_UP, LPM_DOWN);
    type Arith_Type	is (LPM_NO_TYP, LPM_ADD, LPM_SUB);
    type Fflop_Type	is (LPM_DFF,LPM_TFF);
    type ShDir_Type	is (LPM_LEFT,LPM_RIGHT);
    type goal_type	is (speed,area,combinatorial);
    type port_state_type is (ps_tied_low,ps_tied_high,ps_open,ps_connected);

    signal zero : std_logic := '0' ;
    signal one : std_logic := '1' ;

    function port_state (arg : std_logic) return port_state_type;
    function connects (x : std_logic; i : integer) return integer;
    function max (a,b : natural) return natural;
    function vectorize (a, size : natural) RETURN std_logic_vector;
    function restring (a : string; w : positive) return std_logic_vector;
component buf
    port(
	x	: in std_logic;		-- Input to Buffer.
	y	: out std_logic		-- Output from Buffer.
    );
end component;
    COMPONENT logico		-- Master Logic Cell.
	PORT (
	    a1	: IN std_logic;	-- Non-invert to AND A.
	    a2	: IN std_logic;	-- Invert to AND A.
	    a3	: IN std_logic;	-- Non-invert to AND A.
	    a4	: IN std_logic;	-- Invert to AND A.
	    a5	: IN std_logic;	-- Non-invert to AND A.
	    a6	: IN std_logic;	-- Invert to AND A.

	    b1	: IN std_logic;	-- Non-invert to AND B.
	    b2	: IN std_logic;	-- Invert to AND B.

	    c1	: IN std_logic;	-- Non-invert to AND C.
	    c2	: IN std_logic;	-- Invert to AND C.

	    d1	: IN std_logic;	-- Non-invert to AND D.
	    d2	: IN std_logic;	-- Invert to AND D.

	    e1	: IN std_logic;	-- Non-invert to AND E.
	    e2	: IN std_logic;	-- Invert to AND E.

	    f1	: IN std_logic;	-- Non-invert to AND F.
	    f2	: IN std_logic;	-- Invert to AND F.
	    f3	: IN std_logic;	-- Non-invert to AND F.
	    f4	: IN std_logic;	-- Invert to AND F.
	    f5	: IN std_logic;	-- Non-invert to AND F.
	    f6	: IN std_logic;	-- Invert to AND F.

	    oz	: OUT std_logic	-- Output of MUX O.
	);
    END COMPONENT;
    component Madd_sub
	generic(lpm_width	: positive;
		lpm_representation : repre_type := lpm_unsigned;
		lpm_direction	: arith_type := lpm_no_typ;
		lpm_hint	: goal_type := speed);
	port(	dataA		: in	std_logic_vector(lpm_width - 1 downto 0);
		dataB		: in	std_logic_vector(lpm_width - 1 downto 0);
		cin		: in	std_logic;
		add_sub		: in	std_logic;
		result		: out	std_logic_vector(lpm_width - 1 downto 0);
		cout		: out	std_logic;
		overflow	: out   std_logic);
    end component;
end Mmult_local ;
 
package body Mmult_local is
    function port_state (arg : std_logic) return port_state_type is
	variable result : port_state_type := ps_connected;
    begin

	result := arg'port_state_att;

	return result;
    end port_state;

    function connects (x : std_logic; i : integer) return integer is
	variable result : integer := i;
    begin
	if (port_state(x) = ps_connected) then
	    result := result - 1;
	end if;
	return (result);
    end connects;

    function max (a,b : natural) return natural is
    begin
	if (a < b) then return b; else return a; end if;
    end max;

end Mmult_local ;
 

library ieee;
use ieee.std_logic_1164.all;

use work.Mmult_local.all;

entity Mmult is
    generic(lpm_widthA	: positive;
	    lpm_widthB	: positive;
	    lpm_widthS	: natural;
	    lpm_widthP	: positive;
	    lpm_representation : repre_type;
	    lpm_hint	: goal_type);
    port(   dataA	: in	std_logic_vector((lpm_widthA-1) downto 0);
	    dataB	: in	std_logic_vector((lpm_widthB-1) downto 0);
	    sum		: in	std_logic_vector((lpm_widthS-1) downto 0);
	    result	: out	std_logic_vector((lpm_widthP-1) downto 0));
end Mmult;

architecture archlpmstd of Mmult is

    constant psize : natural := max(lpm_widthA,lpm_widthB);
    constant asize : natural := max((lpm_widthA+lpm_widthB),lpm_widthS);
    signal a,b  : std_logic_vector ((psize-1) downto 0);
    signal c,s	: bvm (a'range,b'range);
    signal p	: std_logic_vector (((psize*2)-1) downto 0);
    signal a0,b0,x : std_logic_vector ((asize-1) downto 0);
    signal one  : std_logic := '1';
    signal zero : std_logic := '0';

begin

    -- Make the multiplier symmetrical.

    R0: if (lpm_widthA > lpm_widthB) generate
	b((b'high) downto lpm_widthB) <= (OTHERS => '0');
	b((lpm_widthB - 1) downto 0) <= dataB;
	a <= dataA;
    end generate;
    R1: if (lpm_widthA < lpm_widthB) generate
	a((psize-1) downto lpm_widthA) <= (OTHERS => '0');
	a((lpm_widthA - 1) downto 0) <= dataA;
	b <= dataB;
    end generate;
    R2: if (lpm_widthA = lpm_widthB) generate
	a <= dataA;
	b <= dataB;
    end generate;

    CL: for l in 0 to (psize-2) generate
	CM: for m in 0 to (psize-1) generate
	-- Column 0 of the array.
	    C0: if (m = 0) generate
		SUMM: logico port map(a(m),zero,b(l+1),zero,one,zero,
			zero,one,one,zero,one,zero,zero,one,
			a(m+1),zero,b(l),zero,one,zero, s(l,m));
		CRRY: logico port map(a(m),zero,b(l+1),zero,one,zero,
			zero,one,zero,one,zero,one,one,zero,
			a(m+1),zero,b(l),zero,one,zero, c(l,m));
	    end generate;
	-- Column 1 through m-1 of the array.
	    C1: if ((m > 0) and (m < (psize-1))) generate
		C2: if (l < (psize-2)) generate
		SUMM: logico port map(s(l+1,m-1),zero,one,zero,one,zero,
			c(l,m-1),zero,one,c(l,m-1),one,c(l,m-1),c(l,m-1),zero,
			a(m+1),zero,b(l),zero,one,zero, s(l,m));
		CRRY: logico port map(s(l+1,m-1),zero,one,zero,one,zero,
			zero,one,c(l,m-1),zero,c(l,m-1),zero,one,zero,
			a(m+1),zero,b(l),zero,one,zero, c(l,m));
		end generate;
		C3: if (l = (psize-2)) generate
		SUMM: logico port map(a(m),zero,b(l+1),zero,one,zero,
			c(l,m-1),zero,one,c(l,m-1),one,c(l,m-1),c(l,m-1),zero,
			a(m+1),zero,b(l),zero,one,zero, s(l,m));
		CRRY: logico port map(a(m),zero,b(l+1),zero,one,zero,
			zero,one,c(l,m-1),zero,c(l,m-1),zero,one,zero,
			a(m+1),zero,b(l),zero,one,zero, c(l,m));
		end generate;
	    end generate;
	-- Column m of the array.
	    C4: if (m = (psize-1)) generate
	-- Special case 2x2 multiplier.
		GS: if ((psize-1) <= 1) generate
		SUMM: logico port map(a(m),zero,b(m),zero,one,zero,
			zero,one,one,zero,one,zero,zero,one,
			c(l,m-1),zero,one,zero,one,zero,s(l,m));
		CRRY: logico port map(a(m),zero,b(m),zero,one,zero,
			zero,one,zero,one,zero,one,one,zero,
			c(l,m-1),zero,one,zero,one,zero, c(l,m));
		end generate;
		GL: if ((psize-1)  > 1) generate
		    C5: if (l = 0) generate
			SUMM: logico port map(s(l+1,m-1),zero,one,zero,one,zero,
			  c(l,m-1),zero,one,c(l,m-1),one,c(l,m-1),c(l,m-1),zero,
			  zero,one,zero,one,zero,one, s(l,m));
			CRRY: logico port map(s(l+1,m-1),zero,one,zero,one,zero,
			  zero,one,c(l,m-1),zero,c(l,m-1),zero,one,zero,
			  zero,one,zero,one,zero,one, c(l,m));
		    end generate;
		    C6: if ((l > 0) and (l < (psize-2))) generate
			SUMM: logico port map(s(l+1,m-1),zero,one,zero,one,zero,
			  c(l,m-1),zero,one,c(l,m-1),one,c(l,m-1),c(l,m-1),zero,
			  c(l-1,m),zero,one,zero,one,zero, s(l,m));
			CRRY: logico port map(s(l+1,m-1),zero,one,zero,one,zero,
			  zero,one,c(l,m-1),zero,c(l,m-1),zero,one,zero,
			  c(l-1,m),zero,one,zero,one,zero, c(l,m));
		    end generate;
		    C7: if (l = (psize-2)) generate
		    SUMM:logico port map(a(m),zero,b(l+1),zero,one,zero,
			  c(l,m-1),zero,one,c(l,m-1),one,c(l,m-1),c(l,m-1),zero,
			  c(l-1,m),zero,one,zero,one,zero, s(l,m));
		    CRRY:logico port map(a(m),zero,b(l+1),zero,one,zero,
			  zero,one,c(l,m-1),zero,c(l,m-1),zero,one,zero,
			  c(l-1,m),zero,one,zero,one,zero, c(l,m));
		    end generate;
		end generate;
	    end generate;
	end generate;
    end generate;

    -- Gather the product.

    p(p'low) <= a(0) and b(0);
    G0: for l in 1 to (psize-1) generate
	p(l) <= s(0,l-1);
    end generate;
    G1: for l in psize to (p'high-1) generate
	p(l) <= s(l-psize,(psize-1));
    end generate;
    G2: if (p'length = 2) generate
	p(1) <= '0';
    end generate;
    G3: if (p'length > 2) generate
	p(p'high) <= c(psize-2,(psize-1));
    end generate;

    -- Align the product and the partial sum.

    GA:if (lpm_widthS > 0) generate
	G4: if (lpm_widthA+lpm_widthB < lpm_widthS) generate
	    a0((asize-1) downto (lpm_widthA+lpm_widthB)) <= (OTHERS => '0');
	    a0((lpm_widthA+lpm_widthB-1) downto 0) <=
					p((lpm_widthA+lpm_widthB-1) downto 0);
	    b0 <= sum;
	end generate;
	G5: if (lpm_widthA+lpm_widthB > lpm_widthS) generate
	    a0 <= p;
	    b0((asize-1) downto (lpm_widthS)) <= (OTHERS => '0');
	    b0(sum'range) <= sum;
	end generate;
	G6: if (lpm_widthA+lpm_widthB = lpm_widthS) generate
	    a0 <= p((lpm_widthA+lpm_widthB-1) downto 0);
	    b0 <= sum;
	end generate;

	U0: Madd_sub generic map (asize,lpm_unsigned,lpm_add,lpm_hint)
			port map (a0,b0,zero,one,x,open,open);
    end generate;

    -- No partial sum.

    GN:if (lpm_widthS = 0) generate
	    x <= p((asize-1) downto 0);
    end generate;

    -- Align the result with the most significant std_logic of the product.

    G7: if (lpm_widthP > asize) generate
	result((lpm_widthP-1) downto asize) <= (OTHERS => '0');
	result((asize-1) downto 0) <= x;
    end generate;
    G8: if (lpm_widthP < asize) generate
	result <= x((asize-1) downto (asize-lpm_widthP));
    end generate;
    G9: if (lpm_widthP = asize) generate
	result <= x;
    end generate;

end archlpmstd;
